# Copyright (c) 2020, yolort team. All rights reserved.

import json
from pathlib import Path
from typing import List, Optional, Union

import numpy as np
from PIL import Image


class AnnotationsConverter:
    """
    Make a MSCOCO JSON format annotations from YOLO format. We first find all the images in that
    directory and then match their corresponding labels. The default setting places labels and
    images in the same directory. You can set the name of label_dir to substitute the path for the
    desired labels if the labels and images are not in the same directory. Our replacement strategy
    is to replace image_dir with label_dir, so please also set the image_dir carefully.

    Args:
        root (string): Root directory of the dataset
        metalabels (string | List[string]): Concrete label names of different classes
        image_dir (string, optional): Name of the path to be replaced if it isn't None,
            default is None
        label_dir (string, optional): Name of the replaced path for the desired labels,
            default is None
        split (string, optional): The dataset split, either 'train' (default) or 'test'
        year (int, optional): The year of the dataset, default is 2017
        set_license (bool, optional): Determine whether to set license, default is False
        image_posix (string, optional): Posix of the image, default is 'jpg'
    """

    def __init__(
        self,
        root: str,
        metalabels: Union[str, List[str]],
        image_dir: Optional[str] = None,
        label_dir: Optional[str] = None,
        split: str = "train",
        year: int = 2017,
        set_license: bool = False,
        image_posix: str = "jpg",
    ) -> None:

        self._year = year
        self.type = "instances"
        self.split = f"{split}{year}"
        self.root_path = Path(root)
        self.image_posix = image_posix
        self.image_dir = image_dir or ""
        self.label_dir = label_dir or ""
        self.annotation_root = self._set_annotation_path()
        self.metadata = self._get_metadata(metalabels)
        self.metainfo = self._set_metainfo()
        self.licenses = self._get_licenses(set_license)
        self.categories = self._set_categories()

    def _set_annotation_path(self):
        annotation_root = self.root_path / "annotations"
        Path(annotation_root).mkdir(parents=True, exist_ok=True)
        return annotation_root

    def _set_metainfo(self):
        return {
            "year": self._year,
            "version": "1.0",
            "description": "For object detection",
            "date_created": f"{self._year}",
        }

    @staticmethod
    def _get_metadata(metalabels: Union[str, List[str]]):
        if isinstance(metalabels, list):
            return metalabels

        if isinstance(metalabels, str):
            return np.loadtxt(metalabels, dtype="str", delimiter="\n")

        raise TypeError(f"path of metalabels of list of strings expected, got {type(metalabels)}")

    @staticmethod
    def _get_licenses(set_license):
        if set_license:
            licenses = [
                {
                    "id": 1,
                    "name": "GNU General Public License v3.0",
                    "url": "https://github.com/zhiqwang/yolov5-rt-stack/blob/main/LICENSE",
                }
            ]
            return licenses

        return None

    def _set_categories(self):
        if isinstance(self.metadata[0], dict):
            return [
                {"id": coco_category["id"], "name": coco_category["name"]} for coco_category in self.metadata
            ]
        elif isinstance(self.metadata[0], str):
            return [{"id": label_id, "name": label_name} for label_id, label_name in enumerate(self.metadata)]
        else:
            raise NotImplementedError("Currently doesn't support this methods.")

    def generate(self, coco_type="instances", annotation_format="bbox"):
        image_paths = sorted(self.root_path.rglob(f"*.{self.image_posix}"))
        images, annotations = self._get_image_annotation_pairs(
            image_paths,
            annotation_format=annotation_format,
        )
        json_data = {
            "info": self.metainfo,
            "images": images,
            "type": self.type,
            "annotations": annotations,
            "categories": self.categories,
        }
        if self.licenses is not None:
            json_data["licenses"] = self.licenses
        output_path = self.annotation_root / f"{coco_type}_{self.split}.json"
        with open(output_path, "w") as json_file:
            json.dump(json_data, json_file, sort_keys=True)

    def _get_image_annotation_pairs(self, image_paths, annotation_format="bbox"):
        images = []
        annotations = []
        annotation_id = 0
        for img_id, img_path in enumerate(image_paths, 1):
            label_path = str(img_path).replace(f"{self.image_posix}", "txt")
            label_path = label_path.replace(self.image_dir, self.label_dir)
            width, height = Image.open(img_path).size

            images.append(
                {
                    "date_captured": f"{self._year}",
                    "file_name": str(Path(img_path).relative_to(self.root_path)),
                    "id": img_id,
                    "license": 1,
                    "url": "",
                    "height": height,
                    "width": width,
                }
            )

            with open(label_path, "r") as f:
                for line in f:
                    label_info = line.strip().split()
                    assert len(label_info) == 5
                    annotation_id += 1

                    category_id, vertex_info = label_info[0], label_info[1:]
                    category_id = self.categories[int(category_id)]["id"]
                    if annotation_format == "bbox":
                        segmentation, bbox, area = self._get_annotation(vertex_info, height, width)
                    else:
                        raise NotImplementedError

                    annotations.append(
                        {
                            "segmentation": segmentation,
                            "area": area,
                            "iscrowd": 0,
                            "image_id": img_id,
                            "bbox": bbox,
                            "category_id": category_id,
                            "id": annotation_id,
                        }
                    )

        return images, annotations

    @staticmethod
    def _get_annotation(vertex_info, height, width):

        cx, cy, w, h = [float(i) for i in vertex_info]
        cx = cx * width
        cy = cy * height
        w = w * width
        h = h * height
        x = cx - w / 2
        y = cy - h / 2

        segmentation = [[x, y, x + w, y, x + w, y + h, x, y + h]]
        area = w * h

        bbox = [x, y, w, h]
        return segmentation, bbox, area



"""
Test for exporting model to ONNX and inference with ONNX Runtime
"""
import io
from pathlib import Path

import pytest
import torch
from torch import Tensor
from torchvision.io import read_image
from torchvision.ops._register_onnx_ops import _onnx_opset_version
from yolort import models
from yolort.runtime import PredictorORT
from yolort.runtime.ort_helper import export_onnx
from yolort.utils.image_utils import to_numpy

# In environments without onnxruntime we prefer to
# invoke all tests in the repo and have this one skipped rather than fail.
onnxruntime = pytest.importorskip("onnxruntime")


class TestONNXExporter:
    def run_model(self, model, inputs_list):
        """
        The core part of exporting model to ONNX and inference with ONNX Runtime
        Copy-paste from <https://github.com/pytorch/vision/blob/07fb8ba/test/test_onnx.py#L34>
        """
        model = model.eval()

        onnx_io = io.BytesIO()

        # export to onnx models
        batch_size = len(inputs_list[0])
        export_onnx(onnx_io, model=model, opset_version=_onnx_opset_version, batch_size=batch_size)

        # validate the exported model with onnx runtime
        for test_inputs in inputs_list:
            with torch.no_grad():
                if isinstance(test_inputs, Tensor) or isinstance(test_inputs, list):
                    test_inputs = (test_inputs,)
                test_outputs = model(*test_inputs)
                if isinstance(test_outputs, Tensor):
                    test_outputs = (test_outputs,)
            self.ort_validate(onnx_io, test_inputs, test_outputs)

    def ort_validate(self, onnx_io, inputs, outputs):

        inputs, _ = torch.jit._flatten(inputs)
        outputs, _ = torch.jit._flatten(outputs)

        inputs = list(map(to_numpy, inputs))
        outputs = list(map(to_numpy, outputs))

        y_runtime = PredictorORT(onnx_io.getvalue())
        # Inference on ONNX Runtime
        ort_outs = y_runtime.predict(inputs)

        for i in range(0, len(outputs)):
            torch.testing.assert_allclose(outputs[i], ort_outs[i], rtol=1e-03, atol=1e-05)

    def get_image(self, img_name):

        img_path = Path(__file__).parent.resolve() / "assets" / img_name
        image = read_image(str(img_path)) / 255

        return image

    def get_test_images(self):
        return self.get_image("bus.jpg"), self.get_image("zidane.jpg")

    @pytest.mark.parametrize(
        "arch, fixed_size, upstream_version",
        [
            ("yolov5s", True, "r3.1"),
            ("yolov5m", False, "r4.0"),
            ("yolov5n", True, "r6.0"),
            ("yolov5n", False, "r6.0"),
            ("yolov5n6", False, "r6.0"),
        ],
    )
    def test_onnx_export_single_image(self, arch, fixed_size, upstream_version):
        img_one, img_two = self.get_test_images()
        img_dummy = torch.ones(3, 1080, 720) * 0.3

        size = (640, 640) if arch[-1] == "6" else (320, 320)
        model = models.__dict__[arch](
            upstream_version=upstream_version,
            pretrained=True,
            size=size,
            fixed_shape=size if fixed_size else None,
            score_thresh=0.45,
        )
        model = model.eval()
        model([img_one])
        # Test exported model on images of different size, or dummy input
        self.run_model(model, [[img_one], [img_two], [img_dummy]])

    @pytest.mark.parametrize("arch", ["yolov5n6"])
    def test_onnx_export_multi_batches(self, arch):
        img_one, img_two = self.get_test_images()
        img_dummy = torch.ones(3, 1080, 720) * 0.3

        size = (640, 640) if arch[-1] == "6" else (320, 320)
        model = models.__dict__[arch](pretrained=True, size=size, score_thresh=0.45)
        model = model.eval()
        model([img_one, img_two])

        # Test exported model on images of different size, or dummy input
        inputs_list = [
            [img_one, img_two],
            [img_two, img_one],
            [img_dummy, img_one],
            [img_one, img_one],
            [img_two, img_dummy],
            [img_dummy, img_two],
        ]
        self.run_model(model, inputs_list)

    @pytest.mark.parametrize("arch", ["yolov5n"])
    def test_onnx_export_misbatch(self, arch):
        img_one, img_two = self.get_test_images()
        img_dummy = torch.ones(3, 640, 480) * 0.3

        size = (640, 640) if arch[-1] == "6" else (320, 320)
        model = models.__dict__[arch](pretrained=True, size=size, score_thresh=0.45)
        model = model.eval()
        model([img_one, img_two])

        # Test exported model on images of misbatch
        with pytest.raises(IndexError, match="list index out of range"):
            self.run_model(model, [[img_one, img_two], [img_two, img_one, img_dummy]])

        # Test exported model on images of misbatch
        with pytest.raises(ValueError, match="Model requires 3 inputs. Input Feed contains 2"):
            self.run_model(model, [[img_two, img_one, img_dummy], [img_one, img_two]])



# Copyright (c) 2022, yolort team. All rights reserved.

import logging
from collections import OrderedDict, namedtuple
from typing import Any, Dict, List, Callable, Optional, Tuple

# import yolort before pytorch: this may cause issue in python inference on windows
import yolort.utils.dependency as _dependency

if _dependency.is_module_available("tensorrt"):
    import tensorrt as trt

import numpy as np
import torch
from torch import Tensor
from torchvision.io import read_image
from yolort.models.transform import YOLOTransform
from yolort.utils import contains_any_tensor


logging.basicConfig(level=logging.INFO)
logging.getLogger("PredictorTRT").setLevel(logging.INFO)
logger = logging.getLogger("PredictorTRT")


class PredictorTRT:
    """
    Create a simple end-to-end predictor with the given checkpoint that runs on
    single device for input images with TensorRT.

    Args:
        engine_path (string): Path of the serialized TensorRT engine.
        device (torch.device): The CUDA device to be used for inferencing.
        precision (string): The datatype to use for the engine, either 'fp32', 'fp16' or 'int8'.
        enable_dynamic (bool): Whether to specify axes of tensors as dynamic. Default: False.
        size: (Tuple[int, int]): the minimum and maximum size of the image to be rescaled.
            Default: (640, 640)
        size_divisible (int): stride of the models. Default: 32
        fixed_shape (Tuple[int, int], optional): Padding mode for letterboxing. If set to `True`,
            the image will be padded to shape `fixed_shape` if specified. Instead the image will
            be padded to a minimum rectangle to match `min_size / max_size` and each of its edges
            is divisible by `size_divisible` if it is not specified. Default: None
        fill_color (int): fill value for padding. Default: 114

    Example:

        Demo pipeline for deploying TensorRT.

        .. code-block:: python

            import torch
            from yolort.runtime import PredictorTRT

            # Load the exported TensorRT engine
            engine_path = 'yolov5n6.engine'
            device = torch.device('cuda')
            y_runtime = PredictorTRT(engine_path, device=device)

            # Perform inference on an image file
            predictions = y_runtime.predict('bus.jpg')
    """

    def __init__(
        self,
        engine_path: str,
        device: torch.device = torch.device("cuda"),
        precision: str = "fp32",
        enable_dynamic: bool = False,
        size: Tuple[int, int] = (640, 640),
        size_divisible: int = 32,
        fixed_shape: Optional[Tuple[int, int]] = None,
        fill_color: int = 114,
    ) -> None:

        self._engine_path = engine_path
        self._device = device

        # Build the inference engine
        self.named_binding = namedtuple("Binding", ("name", "dtype", "shape", "data", "ptr"))
        self.engine = self._build_engine()
        self.bindings = OrderedDict()
        self.binding_addrs = None
        self.context = None
        self._set_context()

        if precision == "fp32":
            self._half = False
        elif precision == "fp16":
            self._half = True
        else:
            raise NotImplementedError(f"Currently not supports precision: {precision}")

        self._dtype = torch.float16 if self._half else torch.float32

        # Set pre-processing transform for TensorRT inference
        self._enable_dynamic = enable_dynamic
        self._size = size
        self._size_divisible = size_divisible
        self._fixed_shape = fixed_shape
        self._fill_color = fill_color
        self._img_size = None
        self.transform = None
        self._set_preprocessing()

        # Visualization
        self._names = [f"class{i}" for i in range(1000)]  # assign defaults

    @_dependency.requires_module("tensorrt")
    def _build_engine(self):
        logger.info(f"Loading {self._engine_path} for TensorRT inference...")
        trt_logger = trt.Logger(trt.Logger.INFO)

        trt.init_libnvinfer_plugins(trt_logger, namespace="")
        with open(self._engine_path, "rb") as f, trt.Runtime(trt_logger) as runtime:
            engine = runtime.deserialize_cuda_engine(f.read())

        return engine

    @_dependency.requires_module("tensorrt")
    def _set_context(self):
        for index in range(self.engine.num_bindings):
            name = self.engine.get_binding_name(index)
            dtype = trt.nptype(self.engine.get_binding_dtype(index))
            shape = tuple(self.engine.get_binding_shape(index))
            data = torch.from_numpy(np.empty(shape, dtype=np.dtype(dtype))).to(self._device)
            self.bindings[name] = self.named_binding(name, dtype, shape, data, int(data.data_ptr()))

        self.binding_addrs = OrderedDict((n, d.ptr) for n, d in self.bindings.items())
        self.context = self.engine.create_execution_context()

    def _set_preprocessing(self):
        if self._enable_dynamic:
            raise NotImplementedError("Currently only supports static shape inference in TensorRT.")

        export_onnx_shape = self.bindings["images"].shape
        self._img_size = export_onnx_shape

        size = export_onnx_shape[-2:]
        self.transform = YOLOTransform(
            size[0],
            size[1],
            size_divisible=self._size_divisible,
            fixed_shape=size,
            fill_color=self._fill_color,
        )

    def warmup(self):
        # Warmup model by running inference once and only warmup GPU models
        if isinstance(self._device, torch.device) and self._device.type != "cpu":
            image = torch.zeros(*self._img_size).to(dtype=self._dtype, device=self._device)
            self(image)

    def __call__(self, image: Tensor):
        """
        Args:
            image (Tensor): an image of shape (N, C, H, W).

        Returns:
            predictions (Tuple[Tensor, Tensor, Tensor, Tensor]):
                stands for boxes, scores, labels and number of boxes respectively.
        """
        self.binding_addrs["images"] = int(image.data_ptr())
        self.context.execute_v2(list(self.binding_addrs.values()))
        num_dets = self.bindings["num_detections"].data
        boxes = self.bindings["detection_boxes"].data
        scores = self.bindings["detection_scores"].data
        labels = self.bindings["detection_classes"].data
        return boxes, scores, labels, num_dets

    @staticmethod
    def parse_output(all_boxes, all_scores, all_labels, all_num_dets):
        detections: List[Dict[str, Tensor]] = []

        for boxes, scores, labels, num_dets in zip(all_boxes, all_scores, all_labels, all_num_dets):
            keep = num_dets.item()
            boxes, scores, labels = boxes[:keep], scores[:keep], labels[:keep]
            detections.append({"scores": scores, "labels": labels, "boxes": boxes})

        return detections

    def forward(self, inputs: List[Tensor]):
        """
        Wrapper the TensorRT inference engine with Pre-Processing Module.

        Args:
            inputs (list[Tensor]): images to be processed
        """
        # get the original image sizes
        original_image_sizes: List[Tuple[int, int]] = []

        for img in inputs:
            val = img.shape[-2:]
            assert len(val) == 2
            original_image_sizes.append((val[0], val[1]))

        # Pre-Processing
        samples, _ = self.transform(inputs)
        # Inference on TensorRT
        boxes, scores, labels, num_dets = self(samples.tensors)
        results = self.parse_output(boxes, scores, labels, num_dets)

        # Rescale coordinate
        im_shape = torch.tensor(samples.tensors.shape[-2:])
        detections = self.transform.postprocess(results, im_shape, original_image_sizes)

        return detections

    def predict(self, x: Any, image_loader: Optional[Callable] = None) -> List[Dict[str, Tensor]]:
        """
        Predict function for raw data or processed data
        Args:
            x: Input to predict. Can be raw data or processed data.
            image_loader: Utility function to convert raw data to Tensor.

        Returns:
            The post-processed model predictions.
        """
        image_loader = image_loader or self.default_loader
        images = self.collate_images(x, image_loader)
        return self.forward(images)

    def default_loader(self, img_path: str) -> Tensor:
        """
        Default loader of read a image path.

        Args:
            img_path (str): a image path

        Returns:
            Tensor, processed tensor for prediction.
        """
        return read_image(img_path) / 255.0

    def collate_images(self, samples: Any, image_loader: Callable) -> List[Tensor]:
        """
        Prepare source samples for inference.

        Args:
            samples (Any): samples source, support the following various types:
                - str or List[str]: a image path or list of image paths.
                - Tensor or List[Tensor]: a tensor or list of tensors.

        Returns:
            List[Tensor], The processed image samples.
        """
        if isinstance(samples, Tensor):
            return [samples.to(dtype=self._dtype, device=self._device)]

        if contains_any_tensor(samples):
            return [sample.to(dtype=self._dtype, device=self._device) for sample in samples]

        if isinstance(samples, str):
            samples = [samples]

        if isinstance(samples, (list, tuple)) and all(isinstance(p, str) for p in samples):
            outputs = []
            for sample in samples:
                output = image_loader(sample).to(dtype=self._dtype, device=self._device)
                outputs.append(output)
            return outputs

        raise NotImplementedError(
            f"The type of the sample is {type(samples)}, we currently don't support it now, the "
            "samples should be either a tensor, list of tensors, a image path or list of image paths."
        )



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Dataloaders and dataset utils
"""
import glob
import os
from pathlib import Path

import cv2
import numpy as np
from PIL import Image

from .augmentations import letterbox


# Parameters

# acceptable image suffixes
IMG_FORMATS = ["bmp", "jpg", "jpeg", "png", "tif", "tiff", "dng", "webp", "mpo"]
# acceptable video suffixes
VID_FORMATS = ["mov", "avi", "mp4", "mpg", "mpeg", "m4v", "wmv", "mkv"]
WORLD_SIZE = int(os.getenv("WORLD_SIZE", 1))  # DPP


def exif_transpose(image):
    """
    Transpose a PIL image accordingly if it has an EXIF Orientation tag.
    Inplace exif_transpose() version of
    https://github.com/python-pillow/Pillow/blob/master/src/PIL/ImageOps.py

    Args:
        image: The image to transpose.

    Return:
        An image.
    """
    exif = image.getexif()
    orientation = exif.get(0x0112, 1)  # default 1
    if orientation > 1:
        method = {
            2: Image.FLIP_LEFT_RIGHT,
            3: Image.ROTATE_180,
            4: Image.FLIP_TOP_BOTTOM,
            5: Image.TRANSPOSE,
            6: Image.ROTATE_270,
            7: Image.TRANSVERSE,
            8: Image.ROTATE_90,
        }.get(orientation)
        if method is not None:
            image = image.transpose(method)
            del exif[0x0112]
            image.info["exif"] = exif.tobytes()
    return image


class LoadImages:
    """
    YOLOv5 image/video dataloader. And we're using th CHW RGB format.
    """

    def __init__(self, path: str, img_size: int = 640, stride: int = 32, auto: bool = True):
        path_source = str(Path(path).resolve())  # os-agnostic absolute path
        if "*" in path_source:
            files = sorted(glob.glob(path_source, recursive=True))  # glob
        elif os.path.isdir(path_source):
            files = sorted(glob.glob(os.path.join(path_source, "*.*")))  # dir
        elif os.path.isfile(path_source):
            files = [path_source]  # files
        else:
            raise Exception(f"ERROR: {path_source} does not exist")

        images = [x for x in files if x.split(".")[-1].lower() in IMG_FORMATS]
        videos = [x for x in files if x.split(".")[-1].lower() in VID_FORMATS]
        num_images, num_videos = len(images), len(videos)

        self.img_size = img_size
        self.stride = stride
        self.files = images + videos
        self.num_files = num_images + num_videos  # number of files
        self.video_flag = [False] * num_images + [True] * num_videos
        self.mode = "image"
        self.auto = auto
        if any(videos):
            self.new_video(videos[0])  # new video
        else:
            self.cap = None
        assert self.num_files > 0, (
            f"No images or videos found in {path_source}. Supported formats "
            f"are:\nimages: {IMG_FORMATS}\nvideos: {VID_FORMATS}"
        )

    def __iter__(self):
        self.count = 0
        return self

    def __next__(self):
        if self.count == self.num_files:
            raise StopIteration
        path = self.files[self.count]

        if self.video_flag[self.count]:
            # Read video
            self.mode = "video"
            ret_val, img_origin = self.cap.read()
            while not ret_val:
                self.count += 1
                self.cap.release()
                if self.count == self.num_files:  # last video
                    raise StopIteration
                else:
                    path = self.files[self.count]
                    self.new_video(path)
                    ret_val, img_origin = self.cap.read()

            self.frame += 1
            source_bar = f"video {self.count + 1}/{self.num_files} ({self.frame}/{self.frames}) {path}: "

        else:
            # Read image
            self.count += 1
            img_origin = cv2.imread(path)  # opencv set the BGR order as the default
            assert img_origin is not None, f"Not Found Image: {path}"
            source_bar = f"image {self.count}/{self.num_files} {path}: "

        # Padded resize
        img = letterbox(img_origin, self.img_size, stride=self.stride, auto=self.auto)[0]

        # Convert HWC to CHW, BGR to RGB
        img = img.transpose((2, 0, 1))[::-1]
        img = np.ascontiguousarray(img)

        return path, img, img_origin, self.cap, source_bar

    def new_video(self, path):
        self.frame = 0
        self.cap = cv2.VideoCapture(path)
        self.frames = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))

    def __len__(self):
        return self.num_files  # number of files



# Copyright (c) 2020, yolort team. All rights reserved.

from functools import reduce
from typing import Dict, List, Optional

import torch
from torch import nn
from yolort.v5 import get_yolov5_size, load_yolov5_model

from .backbone_utils import darknet_pan_backbone
from .box_head import YOLOHead

__all__ = ["convert_yolov5_checkpoint", "load_from_ultralytics"]


def load_from_ultralytics(checkpoint_path: str, version: str = "r6.0"):
    """
    Allows the user to load model state file from the checkpoint trained from
    the ultralytics/yolov5.

    Args:
        checkpoint_path (str): Path of the YOLOv5 checkpoint model.
        version (str): upstream version released by the ultralytics/yolov5, Possible
            values are ["r3.1", "r4.0", "r6.0"]. Default: "r6.0".
    """

    if version not in ["r3.1", "r4.0", "r6.0"]:
        raise NotImplementedError(
            f"Currently does not support version: {version}. Feel free to file an issue "
            "labeled enhancement to us."
        )

    checkpoint_yolov5 = load_yolov5_model(checkpoint_path)
    num_classes = checkpoint_yolov5.yaml["nc"]
    strides = checkpoint_yolov5.stride
    # YOLOv5 will change the anchors setting when using the auto-anchor mechanism. So we
    # use the following formula to compute the anchor_grids instead of attaching it via
    # checkpoint_yolov5.yaml["anchors"]
    num_anchors = checkpoint_yolov5.model[-1].anchors.shape[1]
    anchor_grids = (
        (checkpoint_yolov5.model[-1].anchors * checkpoint_yolov5.model[-1].stride.view(-1, 1, 1))
        .reshape(1, -1, 2 * num_anchors)
        .tolist()[0]
    )

    depth_multiple = checkpoint_yolov5.yaml["depth_multiple"]
    width_multiple = checkpoint_yolov5.yaml["width_multiple"]

    use_p6 = False
    if len(strides) == 4:
        use_p6 = True

    if use_p6:
        inner_block_maps = {"0": "11", "1": "12", "3": "15", "4": "16", "6": "19", "7": "20"}
        layer_block_maps = {"0": "23", "1": "24", "2": "26", "3": "27", "4": "29", "5": "30", "6": "32"}
        p6_block_maps = {"0": "9", "1": "10"}
        head_ind = 33
        head_name = "m"
    else:
        inner_block_maps = {"0": "9", "1": "10", "3": "13", "4": "14"}
        layer_block_maps = {"0": "17", "1": "18", "2": "20", "3": "21", "4": "23"}
        p6_block_maps = None
        head_ind = 24
        head_name = "m"

    convert_yolo_checkpoint = CheckpointConverter(
        depth_multiple,
        width_multiple,
        inner_block_maps=inner_block_maps,
        layer_block_maps=layer_block_maps,
        p6_block_maps=p6_block_maps,
        strides=strides,
        anchor_grids=anchor_grids,
        head_ind=head_ind,
        head_name=head_name,
        num_classes=num_classes,
        version=version,
        use_p6=use_p6,
    )
    convert_yolo_checkpoint.updating(checkpoint_yolov5)
    state_dict = convert_yolo_checkpoint.model.half().state_dict()

    size = get_yolov5_size(depth_multiple, width_multiple)

    return {
        "num_classes": num_classes,
        "depth_multiple": depth_multiple,
        "width_multiple": width_multiple,
        "strides": strides,
        "anchor_grids": anchor_grids,
        "use_p6": use_p6,
        "size": size,
        "state_dict": state_dict,
    }


def convert_yolov5_checkpoint(
    checkpoint_path: str,
    output_path: str,
    version: str = "r6.0",
    prefix: str = "yolov5_darknet_pan",
    postfix: str = "custom.pt",
):
    """
    Convert model checkpoint trained with ultralytics/yolov5 to yolort.

    Args:
        checkpoint_path (str): Path of the YOLOv5 checkpoint model.
        output_path (str): Path of the converted yolort checkpoint model.
        version (str): upstream version released by the ultralytics/yolov5, Possible
            values are ["r3.1", "r4.0", "r6.0"]. Default: "r6.0".
        prefix (str): The prefix string of the saved model. Default: "yolov5_darknet_pan".
        postfix (str): The postfix string of the saved model. Default: "custom.pt".
    """

    model_info = load_from_ultralytics(checkpoint_path, version=version)
    model_state_dict = model_info["state_dict"]

    size = model_info["size"]
    use_p6 = "6" if model_info["use_p6"] else ""
    output_postfix = f"{prefix}_{size}{use_p6}_{version.replace('.', '')}_{postfix}"
    torch.save(model_state_dict, output_path / output_postfix)


class ModelWrapper(nn.Module):
    def __init__(self, backbone, head):
        super().__init__()
        self.backbone = backbone
        self.head = head


class CheckpointConverter:
    """
    Update checkpoint from ultralytics yolov5.
    """

    def __init__(
        self,
        depth_multiple: float,
        width_multiple: float,
        inner_block_maps: Optional[Dict[str, str]] = None,
        layer_block_maps: Optional[Dict[str, str]] = None,
        p6_block_maps: Optional[Dict[str, str]] = None,
        strides: Optional[List[int]] = None,
        anchor_grids: Optional[List[List[float]]] = None,
        head_ind: int = 24,
        head_name: str = "m",
        num_classes: int = 80,
        version: str = "r6.0",
        use_p6: bool = False,
    ) -> None:

        # Configuration for making the keys consistent
        if inner_block_maps is None:
            inner_block_maps = {"0": "9", "1": "10", "3": "13", "4": "14"}
        self.inner_block_maps = inner_block_maps
        if layer_block_maps is None:
            layer_block_maps = {"0": "17", "1": "18", "2": "20", "3": "21", "4": "23"}
        self.layer_block_maps = layer_block_maps
        self.p6_block_maps = p6_block_maps
        self.head_ind = head_ind
        self.head_name = head_name

        # Set model
        yolov5_size = get_yolov5_size(depth_multiple, width_multiple)
        backbone_name = f"darknet_{yolov5_size}_{version.replace('.', '_')}"

        backbone = darknet_pan_backbone(
            backbone_name, depth_multiple, width_multiple, version=version, use_p6=use_p6
        )
        num_anchors = len(anchor_grids[0]) // 2
        head = YOLOHead(backbone.out_channels, num_anchors, strides, num_classes)
        # Only backbone and head contain parameters inside, so we only wrap them both here.
        self.model = ModelWrapper(backbone, head)

    def updating(self, state_dict):
        # Obtain module state
        state_dict = obtain_module_sequential(state_dict)

        # Update backbone weights
        for name, params in self.model.backbone.body.named_parameters():
            params.data.copy_(self.attach_parameters_block(state_dict, name, None))

        for name, buffers in self.model.backbone.body.named_buffers():
            buffers.copy_(self.attach_parameters_block(state_dict, name, None))

        # Update PAN weights
        # Updating P6 weights
        if self.p6_block_maps is not None:
            for name, params in self.model.backbone.pan.intermediate_blocks.p6.named_parameters():
                params.data.copy_(self.attach_parameters_block(state_dict, name, self.p6_block_maps))

            for name, buffers in self.model.backbone.pan.intermediate_blocks.p6.named_buffers():
                buffers.copy_(self.attach_parameters_block(state_dict, name, self.p6_block_maps))

        # Updating inner_block weights
        for name, params in self.model.backbone.pan.inner_blocks.named_parameters():
            params.data.copy_(self.attach_parameters_block(state_dict, name, self.inner_block_maps))

        for name, buffers in self.model.backbone.pan.inner_blocks.named_buffers():
            buffers.copy_(self.attach_parameters_block(state_dict, name, self.inner_block_maps))

        # Updating layer_block weights
        for name, params in self.model.backbone.pan.layer_blocks.named_parameters():
            params.data.copy_(self.attach_parameters_block(state_dict, name, self.layer_block_maps))

        for name, buffers in self.model.backbone.pan.layer_blocks.named_buffers():
            buffers.copy_(self.attach_parameters_block(state_dict, name, self.layer_block_maps))

        # Update YOLOHead weights
        for name, params in self.model.head.named_parameters():
            params.data.copy_(self.attach_parameters_heads(state_dict, name))

        for name, buffers in self.model.head.named_buffers():
            buffers.copy_(self.attach_parameters_heads(state_dict, name))

    @staticmethod
    def attach_parameters_block(state_dict, name, block_maps=None):
        keys = name.split(".")
        ind = int(block_maps[keys[0]]) if block_maps else int(keys[0])
        return rgetattr(state_dict[ind], keys[1:])

    def attach_parameters_heads(self, state_dict, name):
        keys = name.split(".")
        ind = int(keys[1])
        return rgetattr(getattr(state_dict[self.head_ind], self.head_name)[ind], keys[2:])


def rgetattr(obj, attr, *args):
    """
    Nested version of getattr.
    Ref: https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-objects
    """

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj] + attr)


def obtain_module_sequential(state_dict):
    if isinstance(state_dict, nn.Sequential):
        return state_dict
    else:
        return obtain_module_sequential(state_dict.model)



# Copyright (c) 2021, yolort team. All rights reserved.
#
# This source code is licensed under the GPL-3.0 license found in the
# LICENSE file in the root directory of this source tree.
#
# Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
#
# This source code is licensed under the Apache-2.0 license found in the
# LICENSE file in the root directory of TensorRT source tree.
#

import logging
from pathlib import Path
from typing import Optional

import torch
import yolort.utils.dependency as _dependency
from torch import Tensor

if _dependency.is_module_available("tensorrt"):
    import tensorrt as trt

from yolort.relay.trt_graphsurgeon import YOLOTRTGraphSurgeon

logging.basicConfig(level=logging.INFO)
logging.getLogger("TRTHelper").setLevel(logging.INFO)
logger = logging.getLogger("TRTHelper")


def export_tensorrt_engine(
    checkpoint_path,
    *,
    score_thresh: float = 0.25,
    nms_thresh: float = 0.45,
    version: str = "r6.0",
    onnx_path: Optional[str] = None,
    engine_path: Optional[str] = None,
    input_sample: Optional[Tensor] = None,
    detections_per_img: int = 100,
    precision: str = "fp32",
    verbose: bool = False,
    workspace: int = 12,
) -> None:
    """
    Export ONNX models and TensorRT serialized engines that can be used for TensorRT inferencing.

    Args:
        checkpoint_path (str): Path of the YOLOv5 checkpoint model.
        score_thresh (float): Score threshold used for postprocessing the detections. Default: 0.25
        nms_thresh (float): NMS threshold used for postprocessing the detections. Default: 0.45
        version (str): upstream version released by the ultralytics/yolov5, Possible
            values are ["r3.1", "r4.0", "r6.0"]. Default: "r6.0".
        onnx_path (string, optional): The path to the ONNX graph to load. Default: None
        engine_path (string, optional): The path where to serialize the engine to. Default: None
        input_sample (Tensor, optional): Specify the input shape to export ONNX, and the
            default shape for the sample is (1, 3, 640, 640).
        detections_per_img (int): Number of best detections to keep after NMS. Default: 100
        precision (string): The datatype to use for the engine inference, either 'fp32', 'fp16' or
            'int8'. Default: 'fp32'
        verbose (bool): If enabled, a higher verbosity level will be set on the TensorRT
            logger. Default: False
        workspace (int): Max memory workspace to allow, in Gb. Default: 12
    """

    if input_sample is None:
        input_sample = torch.rand(1, 3, 640, 640)

    yolo_gs = YOLOTRTGraphSurgeon(checkpoint_path, version=version, input_sample=input_sample)

    # Register the `EfficientNMS_TRT` into the graph.
    yolo_gs.register_nms(
        score_thresh=score_thresh,
        nms_thresh=nms_thresh,
        detections_per_img=detections_per_img,
    )

    # Set the path of ONNX and Tensorrt Engine to export
    checkpoint_path = Path(checkpoint_path)
    onnx_path = onnx_path or str(checkpoint_path.with_suffix(".onnx"))
    engine_path = engine_path or str(checkpoint_path.with_suffix(".engine"))

    # Save the exported ONNX models.
    yolo_gs.save(onnx_path)

    # Build and export the TensorRT engine.
    engine_builder = EngineBuilder(verbose=verbose, workspace=workspace, precision=precision)
    engine_builder.create_network(onnx_path)
    engine_builder.create_engine(engine_path)


class EngineBuilder:
    """
    Parses an ONNX graph and builds a TensorRT engine from it.
    """

    @_dependency.requires_module("tensorrt")
    def __init__(
        self,
        verbose: bool = False,
        workspace: int = 4,
        precision: str = "fp32",
        enable_dynamic: bool = False,
        max_batch_size: int = 16,
        calib_input: Optional[str] = None,
        calib_cache: Optional[str] = None,
        calib_num_images: int = 5000,
        calib_batch_size: int = 8,
    ):
        """
        Args:
            verbose (bool): If enabled, a higher verbosity level will be set on the TensorRT
                logger. Default: False
            workspace (int): Max memory workspace to allow, in Gb. Default: 4
            precision (string): The datatype to use for the engine inference, either 'fp32',
                'fp16' or 'int8'. Default: 'fp32'
            enable_dynamic (bool): Whether to enable dynamic shapes. Default: False
            max_batch_size (int): Maximum batch size reserved for dynamic shape inference.
                Default: 16
            calib_input (string, optinal): The path to a directory holding the calibration images.
                Default: None
            calib_cache (string, optinal): The path where to write the calibration cache to,
                or if it already exists, load it from. Default: None
            calib_num_images (int): The maximum number of images to use for calibration. Default: 5000
            calib_batch_size (int): The batch size to use for the calibration process. Default: 8
        """
        self.logger = trt.Logger(trt.Logger.INFO)
        if verbose:
            self.logger.min_severity = trt.Logger.Severity.VERBOSE

        trt.init_libnvinfer_plugins(self.logger, namespace="")

        self.builder = trt.Builder(self.logger)
        self.config = self.builder.create_builder_config()
        self.config.max_workspace_size = workspace * 1 << 30

        self.batch_size = None
        self.network = None
        self.parser = None

        # Leaving some interfaces and parameters for subsequent use, but we have not yet
        # implemented the following functionality
        self.precision = precision
        self.enable_dynamic = enable_dynamic
        self.max_batch_size = max_batch_size
        self.calib_input = calib_input
        self.calib_cache = calib_cache
        self.calib_num_images = calib_num_images
        self.calib_batch_size = calib_batch_size

    def create_network(self, onnx_path: str):
        """
        Parse the ONNX graph and create the corresponding TensorRT network definition.

        Args:
            onnx_path (string): The path to the ONNX graph to load.
        """

        flag = 1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
        self.network = self.builder.create_network(flag)
        self.parser = trt.OnnxParser(self.network, self.logger)
        if not self.parser.parse_from_file(onnx_path):
            raise RuntimeError(f"Failed to load ONNX file: {onnx_path}")

        inputs = [self.network.get_input(i) for i in range(self.network.num_inputs)]
        outputs = [self.network.get_output(i) for i in range(self.network.num_outputs)]

        logger.info("Network Description")
        for input in inputs:
            self.batch_size = input.shape[0]
            logger.info(f"Input '{input.name}' with shape {input.shape} and dtype {input.dtype}")
        for output in outputs:
            logger.info(f"Output '{output.name}' with shape {output.shape} and dtype {output.dtype}")

    def create_engine(self, engine_path: str):
        """
        Build the TensorRT engine and serialize it to disk.

        Args:
            engine_path (string): The path where to serialize the engine to.
        """
        engine_path = Path(engine_path)
        engine_path.parent.mkdir(parents=True, exist_ok=True)

        precision = self.precision
        logger.info(f"Building {precision} Engine in {engine_path}")

        # Process the batch size and profile
        assert self.batch_size > 0, "Currently only supports static shape."
        self.builder.max_batch_size = self.batch_size

        if precision == "fp16":
            if not self.builder.platform_has_fast_fp16:
                logger.warning("FP16 is not supported natively on this platform/device")
            else:
                self.config.set_flag(trt.BuilderFlag.FP16)
        elif precision == "fp32":
            logger.info("Using fp32 mode.")
        else:
            raise NotImplementedError(f"Currently hasn't been implemented: {precision}.")

        with self.builder.build_engine(self.network, self.config) as engine, open(engine_path, "wb") as f:
            f.write(engine.serialize())
            logger.info(f"Serialize engine success, saved as {engine_path}")



# Copyright (c) 2021, yolort team. All rights reserved.

import logging
from typing import Any, Dict, List, Callable, Optional

import numpy as np
import yolort.utils.dependency as _dependency
from yolort.utils import contains_any_tensor

if _dependency.is_module_available("onnxruntime"):
    import onnxruntime as ort

logger = logging.getLogger(__name__)


class PredictorORT:
    """
    Create a simple end-to-end predictor with the given checkpoint that runs on
    single device for input images with ONNX Runtime.

    Args:
        engine_path (string): Path of the serialized ONNX model.
        device (string): The device to be used for inferencing.

    Example:

        Demo pipeline for deploying yolort with ONNX Runtime.

        .. code-block:: python

            from yolort.runtime import PredictorORT

            # Load the serialized ONNX model
            engine_path = 'yolov5n6.onnx'
            device = 'cpu'
            y_runtime = PredictorORT(engine_path, device=device)

            # Perform inference on an image file
            predictions = y_runtime.predict('bus.jpg')
    """

    def __init__(self, engine_path: str, device: str = "cpu") -> None:
        self.engine_path = engine_path
        self.device = device
        self._providers = self._set_providers()
        self._runtime = self._build_runtime()
        self._input_names = self._runtime.get_inputs()[0].name

    @_dependency.requires_module("onnxruntime")
    def _set_providers(self):
        logger.info("Providers was initialized.")
        ort_device = ort.get_device()
        providers = None

        enable_gpu = True if self.device == "cuda" else False
        if ort_device == "GPU" and enable_gpu:
            providers = ["CUDAExecutionProvider"]
            logger.info("Set inference device to GPU")
        elif enable_gpu:
            logger.info("GPU is not supported by your ONNXRuntime build. Fallback to CPU.")
        else:
            providers = ["CPUExecutionProvider"]
            logger.info("Set inference device to CPU")

        return providers

    @_dependency.requires_module("onnxruntime")
    def _build_runtime(self):
        runtime = ort.InferenceSession(self.engine_path, providers=self._providers)
        return runtime

    def default_loader(self, img_path: str) -> np.ndarray:
        """
        Default loader of read a image path.

        Args:
            img_path (str): the path to the image

        Returns:
            np.ndarray, processed ndarray for prediction.
        """
        import cv2

        image = cv2.imread(img_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = image.astype(np.float32) / 255.0
        image = np.transpose(image, axes=(2, 0, 1))

        return image

    def __call__(self, inputs: List[np.ndarray]):
        """
        Args:
            inputs (List[np.ndarray]): a list of images with shape (C, H, W) (in RGB order).

        Returns:
            predictions (Tuple[List[float], List[int], List[float, float]]):
                stands for scores, labels and boxes respectively.
        """
        inputs = dict((self._runtime.get_inputs()[i].name, inpt) for i, inpt in enumerate(inputs))
        predictions = self._runtime.run(output_names=None, input_feed=inputs)
        return predictions

    def predict(self, x: Any, image_loader: Optional[Callable] = None) -> List[Dict[str, np.ndarray]]:
        """
        Predict function for raw data or processed data

        Args:
            x: Input to predict. Can be raw data or processed data.
            image_loader: Utility function to convert raw data to Numpy ndarray.

        Returns:
            The post-processed model predictions.
        """
        image_loader = image_loader or self.default_loader
        images = self.collate_images(x, image_loader)
        return self(images)

    def collate_images(self, samples: Any, image_loader: Callable) -> List[np.ndarray]:
        """
        Prepare source samples for inference.

        Args:
            samples (Any): samples source, support the following various types:
                - str or List[str]: a image path or list of image paths.
                - np.ndarray or List[np.ndarray]: a ndarray or list of ndarray.

        Returns:
            List[np.ndarray], The processed image samples.
        """
        if isinstance(samples, np.ndarray):
            return [samples]

        if contains_any_tensor(samples, dtype=np.ndarray):
            return [sample for sample in samples]

        if isinstance(samples, str):
            samples = [samples]

        if isinstance(samples, (list, tuple)) and all(isinstance(p, str) for p in samples):
            outputs = []
            for sample in samples:
                output = image_loader(sample)
                outputs.append(output)
            return outputs

        raise NotImplementedError(
            f"The type of the sample is {type(samples)}, we currently don't support it now, the "
            "samples should be either a ndarray, list of np.ndarray, a image path or list of image paths."
        )



# Copyright (c) 2022, yolort team. All rights reserved.

from typing import Optional, Tuple

import torch
from torch import nn
from yolort.models import YOLO, YOLOv5


def export_onnx(
    onnx_path: str,
    *,
    checkpoint_path: Optional[str] = None,
    model: Optional[nn.Module] = None,
    size: Tuple[int, int] = (640, 640),
    size_divisible: int = 32,
    score_thresh: float = 0.25,
    nms_thresh: float = 0.45,
    version: str = "r6.0",
    skip_preprocess: bool = False,
    opset_version: int = 11,
    batch_size: int = 1,
) -> None:
    """
    Export to ONNX models that can be used for ONNX Runtime inferencing.

    Args:
        onnx_path (string): The path to the ONNX graph to be exported.
        checkpoint_path (string, optional): Path of the custom trained YOLOv5 checkpoint.
            Default: None
        model (nn.Module): The defined PyTorch module to be exported. Default: None
        size: (Tuple[int, int]): the minimum and maximum size of the image to be rescaled.
            Default: (640, 640)
        size_divisible (int): Stride in the preprocessing. Default: 32
        score_thresh (float): Score threshold used for postprocessing the detections.
            Default: 0.25
        nms_thresh (float): NMS threshold used for postprocessing the detections. Default: 0.45
        version (string): Upstream YOLOv5 version. Default: 'r6.0'
        skip_preprocess (bool): Skip the preprocessing transformation when exporting the ONNX
            models. Default: False
        opset_version (int): Opset version for exporting ONNX models. Default: 11
        batch_size (int): Only used for models that include pre-processing, you need to specify
            the batch sizes and ensure that the number of input images is the same as the batches
            when inferring if you want to export multiple batches ONNX models. Default: 1
    """

    onnx_builder = ONNXBuilder(
        checkpoint_path=checkpoint_path,
        model=model,
        size=size,
        size_divisible=size_divisible,
        score_thresh=score_thresh,
        nms_thresh=nms_thresh,
        version=version,
        skip_preprocess=skip_preprocess,
        opset_version=opset_version,
        batch_size=batch_size,
    )

    onnx_builder.to_onnx(onnx_path)


class ONNXBuilder:
    """
    YOLOv5 wrapper for exporting ONNX models.

    Args:
        checkpoint_path (string): Path of the custom trained YOLOv5 checkpoint.
        model (nn.Module): The defined PyTorch module to be exported. Default: None
        size: (Tuple[int, int]): the minimum and maximum size of the image to be rescaled.
            Default: (640, 640)
        size_divisible (int): Stride in the preprocessing. Default: 32
        score_thresh (float): Score threshold used for postprocessing the detections.
            Default: 0.25
        nms_thresh (float): NMS threshold used for postprocessing the detections. Default: 0.45
        version (string): Upstream YOLOv5 version. Default: 'r6.0'
        skip_preprocess (bool): Skip the preprocessing transformation when exporting the ONNX
            models. Default: False
        opset_version (int): Opset version for exporting ONNX models. Default: 11
        batch_size (int): Only used for models that include pre-processing, you need to specify
            the batch sizes and ensure that the number of input images is the same as the batches
            when inferring if you want to export multiple batches ONNX models. Default: 1
    """

    def __init__(
        self,
        checkpoint_path: Optional[str] = None,
        model: Optional[nn.Module] = None,
        size: Tuple[int, int] = (640, 640),
        size_divisible: int = 32,
        score_thresh: float = 0.25,
        nms_thresh: float = 0.45,
        version: str = "r6.0",
        skip_preprocess: bool = False,
        opset_version: int = 11,
        batch_size: int = 1,
    ) -> None:

        super().__init__()
        self._checkpoint_path = checkpoint_path
        self._version = version
        # For post-processing
        self._score_thresh = score_thresh
        self._nms_thresh = nms_thresh
        self._skip_preprocess = skip_preprocess
        # For pre-processing
        self._size = size
        self._size_divisible = size_divisible
        self._batch_size = batch_size
        # Define the module
        if model is None:
            model = self._build_model()
        self.model = model

        # For exporting ONNX model
        self._opset_version = opset_version
        self.input_names = self._set_input_names()
        self.output_names = self._set_output_names()
        self.input_sample = self._set_input_sample()
        self.dynamic_axes = self._set_dynamic_axes()

    def _build_model(self):
        if self._skip_preprocess:
            model = YOLO.load_from_yolov5(
                self._checkpoint_path,
                score_thresh=self._score_thresh,
                nms_thresh=self._nms_thresh,
                version=self._version,
            )
        else:
            model = YOLOv5.load_from_yolov5(
                self._checkpoint_path,
                size=self._size,
                size_divisible=self._size_divisible,
                score_thresh=self._score_thresh,
                nms_thresh=self._nms_thresh,
                version=self._version,
            )

        model = model.eval()
        return model

    def _set_input_names(self):
        if self._skip_preprocess:
            return ["images"]
        if self._batch_size == 1:
            return ["image"]

        input_names = []
        for i in range(self._batch_size):
            input_names.append(f"image{i + 1}")
        return input_names

    def _set_output_names(self):
        if self._skip_preprocess:
            return ["scores", "labels", "boxes"]
        if self._batch_size == 1:
            return ["score", "label", "box"]

        output_names = []
        for i in range(self._batch_size):
            output_names.extend([f"score{i + 1}", f"label{i + 1}", f"box{i + 1}"])
        return output_names

    def _set_dynamic_axes(self):
        if self._skip_preprocess:
            return {
                "images": {0: "batch", 2: "height", 3: "width"},
                "boxes": {0: "batch", 1: "num_objects"},
                "labels": {0: "batch", 1: "num_objects"},
                "scores": {0: "batch", 1: "num_objects"},
            }
        if self._batch_size == 1:
            return {
                "image": {1: "height", 2: "width"},
                "box": {0: "num_objects"},
                "label": {0: "num_objects"},
                "score": {0: "num_objects"},
            }

        dynamic_axes = {}
        for i in range(self._batch_size):
            dynamic_axes[f"image{i + 1}"] = {1: "height", 2: "width"}
            dynamic_axes[f"box{i + 1}"] = {0: "num_objects"}
            dynamic_axes[f"label{i + 1}"] = {0: "num_objects"}
            dynamic_axes[f"score{i + 1}"] = {0: "num_objects"}
        return dynamic_axes

    def _set_input_sample(self):
        if self._skip_preprocess:
            return torch.rand(1, 3, 640, 640)
        if self._batch_size == 1:
            return [torch.rand(3, 640, 640)]

        return [torch.rand(3, 640, 640)] * self._batch_size

    @torch.no_grad()
    def to_onnx(self, onnx_path: str, **kwargs):
        """
        Saves the model in ONNX format.

        Args:
            onnx_path (string): The path to the ONNX graph to be exported.
            **kwargs: Will be passed to torch.onnx.export function.
        """

        torch.onnx.export(
            self.model,
            self.input_sample,
            onnx_path,
            do_constant_folding=True,
            opset_version=self._opset_version,
            input_names=self.input_names,
            output_names=self.output_names,
            dynamic_axes=self.dynamic_axes,
            **kwargs,
        )



"""
A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject

Adapted from TorchVision, see:
https://github.com/pytorch/vision/blob/master/setup.py
"""
import os
import subprocess
from pathlib import Path

from setuptools import find_packages, setup

PATH_ROOT = Path(__file__).parent.resolve()

version = "0.7.0a0"
sha = "Unknown"
package_name = "yolort"

try:
    sha = subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=PATH_ROOT).decode("ascii").strip()
except Exception:
    pass

if os.getenv("BUILD_VERSION"):
    version = os.getenv("BUILD_VERSION")
elif sha != "Unknown":
    version += "+" + sha[:7]


def write_version_file():
    version_path = PATH_ROOT / package_name / "version.py"
    with open(version_path, "w") as f:
        f.write(f"__version__ = '{version}'\n")
        f.write(f"git_version = {repr(sha)}\n")
        f.write("from torchvision.extension import _check_cuda_version\n")
        f.write("if _check_cuda_version() > 0:\n")
        f.write("    cuda = _check_cuda_version()\n")


def get_long_description():
    # Get the long description from the README file
    description = (PATH_ROOT / "README.md").read_text(encoding="utf-8")
    # replace relative repository path to absolute link to the release
    static_url = f"https://raw.githubusercontent.com/zhiqwang/yolov5-rt-stack/v{version}"
    description = description.replace("docs/source/_static/", f"{static_url}/docs/source/_static/")
    description = description.replace("notebooks/assets/", f"{static_url}/notebooks/assets/")
    description = description.replace("_graph_visualize.svg", "_graph_visualize.png")
    return description


def load_requirements(path_dir=PATH_ROOT, file_name="requirements.txt", comment_char="#"):
    with open(path_dir / file_name, "r", encoding="utf-8", errors="ignore") as file:
        lines = [ln.rstrip() for ln in file.readlines() if not ln.startswith("#")]
    reqs = []
    for ln in lines:
        if comment_char in ln:  # filer all comments
            ln = ln[: ln.index(comment_char)].strip()
        if ln.startswith("http"):  # skip directly installed dependencies
            continue
        if ln:  # if requirement is not empty
            reqs.append(ln)
    return reqs


if __name__ == "__main__":
    print(f"Building wheel {package_name}-{version}")

    write_version_file()

    setup(
        name=package_name,
        version=version,
        description="yolort is a runtime stack for object detection on specialized accelerators.",
        author="Zhiqiang Wang",
        author_email="me@zhiqwang.com",
        long_description=get_long_description(),
        long_description_content_type="text/markdown",
        url="https://github.com/zhiqwang/yolov5-rt-stack",
        license="GPL-3.0",
        packages=find_packages(exclude=["test", "deployment", "notebooks"]),
        zip_safe=False,
        classifiers=[
            # Operation system
            "Operating System :: OS Independent",
            # How mature is this project? Common values are
            #   3 - Alpha, 4 - Beta, 5 - Production/Stable
            "Development Status :: 4 - Beta",
            # Indicate who your project is intended for
            "Intended Audience :: Developers",
            # Topics
            "Topic :: Education",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Image Recognition",
            # Pick your license as you wish
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            # Specify the Python versions you support here.
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
        ],
        install_requires=load_requirements(),
        # This field adds keywords for your project which will appear on the
        # project page. What does your project relate to?
        #
        # Note that this is a list of additional keywords, separated
        # by commas, to be used to assist searching for the distribution in a
        # larger catalog.
        keywords="machine-learning, deep-learning, ml, pytorch, YOLO, object-detection, YOLOv5, TorchScript",
        # Specify which Python versions you support. In contrast to the
        # 'Programming Language' classifiers above, 'pip install' will check this
        # and refuse to install the project if the version does not match. See
        # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
        python_requires=">=3.6",
        # List additional URLs that are relevant to your project as a dict.
        #
        # This field corresponds to the "Project-URL" metadata fields:
        # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
        #
        # Examples listed include a pattern for specifying where the package tracks
        # issues, where the source is hosted, where to say thanks to the package
        # maintainers, and where to support the project financially. The key is
        # what's used to render the link text on PyPI.
        project_urls={  # Optional
            "Bug Reports": "https://github.com/zhiqwang/yolov5-rt-stack/issues",
            "Funding": "https://zhiqwang.com",
            "Source": "https://github.com/zhiqwang/yolov5-rt-stack/",
        },
    )



# Copyright (c) 2020, yolort team. All rights reserved.

from typing import Any, Dict, Callable, Mapping, Sequence, Type, Union

from torch import Tensor

try:
    from torch.hub import load_state_dict_from_url
except ImportError:
    from torch.utils.model_zoo import load_url as load_state_dict_from_url

from .annotations_converter import AnnotationsConverter
from .dependency import check_version
from .hooks import FeatureExtractor
from .image_utils import cv2_imshow, get_image_from_url, read_image_to_tensor
from .visualizer import Visualizer


__all__ = [
    "AnnotationsConverter",
    "FeatureExtractor",
    "Visualizer",
    "check_version",
    "contains_any_tensor",
    "cv2_imshow",
    "get_image_from_url",
    "get_callable_dict",
    "load_state_dict_from_url",
    "read_image_to_tensor",
]


def get_callable_name(fn_or_class: Union[Callable, object]) -> str:
    return getattr(fn_or_class, "__name__", fn_or_class.__class__.__name__).lower()


def get_callable_dict(fn: Union[Callable, Mapping, Sequence]) -> Union[Dict, Mapping]:
    if isinstance(fn, Mapping):
        return fn
    elif isinstance(fn, Sequence):
        return {get_callable_name(f): f for f in fn}
    elif callable(fn):
        return {get_callable_name(fn): fn}


def contains_any_tensor(value: Any, dtype: Type = Tensor) -> bool:
    """
    Determine whether or not a list contains any Type
    """
    if isinstance(value, dtype):
        return True
    if isinstance(value, (list, tuple)):
        return any(contains_any_tensor(v, dtype=dtype) for v in value)
    elif isinstance(value, dict):
        return any(contains_any_tensor(v, dtype=dtype) for v in value.values())
    return False



# Copyright (c) 2022, yolort team. All rights reserved.

from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
import yolort.utils.dependency as _dependency
from PIL import Image
from torch import Tensor
from yolort.v5.utils.plots import Colors

if _dependency.is_module_available("cv2"):
    import cv2


class Visualizer:
    """
    Visualizer that draws data about detection on images.

    It contains methods like `draw_{text,box}` that draw primitive objects to images, as well as
    high-level wrappers like `draw_{instance_predictions,dataset_dict}` that draw composite data
    in some pre-defined style.

    This visualizer focuses on high rendering quality rather than performance. It is not
    designed to be used for real-time applications.

    Reference:
        We have followed most of the interfaces of detectron2 here, but the implementation will be
        a bit different. Check out the following for more details.
        https://github.com/facebookresearch/detectron2/blob/9258799/detectron2/utils/visualizer.py

    Args:
        image (Tensor | numpy.ndarray): Tensor of shape (C x H x W) or ndarray
            of shape (H x W x C) with dtype uint8.
        metalabels (string | List[string]): Concrete label names of different classes.
    """

    def __init__(self, image: Union[Tensor, np.ndarray], metalabels: Union[str, List[str]]) -> None:
        # Set image and its RGB mode
        self.img, self.is_bgr = self._set_image(image)
        # Set dataset metadata (e.g. class names)
        self.metadata = self._set_metadata(metalabels)
        self.line_width = max(round(sum(self.img.shape) / 2 * 0.003), 2)
        self.assigned_colors = Colors()
        self.output = self.img

    @staticmethod
    def _set_image(image: Union[Tensor, np.ndarray]):
        if isinstance(image, Tensor):
            if image.dtype != torch.uint8:
                raise ValueError(f"Tensor uint8 expected, got {image.dtype}")
            if image.dim() != 3:
                raise ValueError("Pass individual images, not batches")
            if image.size(0) not in {1, 3}:
                raise ValueError("Only grayscale and RGB images are supported")
            # Handle Grayscale images
            if image.size(0) == 1:
                image = torch.tile(image, (3, 1, 1))
            return image.permute(1, 2, 0).cpu().numpy(), False

        if isinstance(image, np.ndarray):
            if image.dtype != np.uint8:
                raise ValueError(f"Numpy uint8 expected, got {image.dtype}")
            if image.ndim != 3:
                raise ValueError("Currently only BGR images are supported")
            return image, True

        raise TypeError(f"Tensor or numpy.ndarray expected, got {type(image)}")

    @staticmethod
    def _set_metadata(metalabels: Union[str, List[str]]):
        if isinstance(metalabels, list):
            return metalabels

        if isinstance(metalabels, str):
            return np.loadtxt(metalabels, dtype="str", delimiter="\n")

        raise TypeError(f"path of metalabels of list of strings expected, got {type(metalabels)}")

    def draw_instance_predictions(self, predictions: Dict[str, Tensor]):
        """
        Draw instance-level prediction results on an image.

        Args:
            predictions (dict): the output of an instance detection model. Following
                fields will be used to draw: "boxes", "labels", "scores".

        Returns:
            np.ndarray: image object with visualizations.
        """

        boxes = self._convert_boxes(predictions["boxes"])
        labels = predictions["labels"].tolist()
        colors = self._create_colors(labels)
        scores = predictions["scores"].tolist()
        labels = self._create_text_labels(labels, scores)

        self.overlay_instances(boxes=boxes, labels=labels, colors=colors)
        return self.output

    @_dependency.requires_module("cv2")
    def imshow(self, scale: Optional[float] = None):
        """
        A replacement of cv2.imshow() for using in Jupyter notebooks.

        Args:
            scale (float, optional): zoom ratio to show the image. Default: None
        """
        from IPython.display import display

        img = self.output

        img = img.clip(0, 255).astype("uint8")
        # cv2 stores colors as BGR; convert to RGB
        if self.is_bgr and img.ndim == 3:
            if img.shape[2] == 4:
                img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)
            else:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        if scale is not None:
            img = cv2.resize(img, None, fx=scale, fy=scale)

        display(Image.fromarray(img))

    def overlay_instances(
        self,
        boxes: np.ndarray,
        labels: List[str],
        colors: List[Tuple[int, int, int]],
    ) -> np.ndarray:
        """
        Overlay bounding boxes and labels on input image.

        Args:
            boxes (np.ndarray): Numpy array of size (N, 4) containing bounding boxes
                in (xmin, ymin, xmax, ymax) format for the N objects in a single image.
                Note that the boxes are absolute coordinates with respect to the image.
                In other words: `0 <= xmin < xmax < W` and `0 <= ymin < ymax < H`.
            labels (List[string]): List containing the text to be displayed for each
                instance.
            colors (List[Tuple[int, int, int]]): List containing the color of the label
                to be painted.

        Returns:
            np.ndarray: image object with visualizations.
        """

        num_instances = len(boxes)
        assert len(labels) == num_instances
        if num_instances == 0:
            return self.output

        # Display in largest to smallest order to reduce occlusion.
        areas = np.prod(boxes[:, 2:] - boxes[:, :2], axis=1)

        sorted_idxs = np.argsort(-areas).tolist()
        # Re-order overlapped instances in descending order.
        boxes = boxes[sorted_idxs] if boxes is not None else None
        labels = [labels[k] for k in sorted_idxs] if labels is not None else None
        colors = [colors[k] for k in sorted_idxs] if colors is not None else None

        for box, label, color in zip(boxes, labels, colors):
            pt1, pt2 = (int(box[0]), int(box[1])), (int(box[2]), int(box[3]))
            self.draw_box(pt1, pt2, color=color)

            lighter_color = self._change_color_brightness(color, brightness_factor=0.7)
            self.draw_text(label, pt1, pt2, color=lighter_color)

        return self.output

    @_dependency.requires_module("cv2")
    def draw_box(
        self,
        pt1: Tuple[int, int],
        pt2: Tuple[int, int],
        color: Tuple[int, int, int] = (229, 160, 21),
    ) -> np.ndarray:
        """
        Draws bounding boxes on given image.
        The values of the input image should be uint8 between 0 and 255.

        Args:
            pt1 (Tuple[int, int]): Vertex of the rectangle (top left corner).
            pt2 (Tuple[int, int]): Vertex of the rectangle opposite to pt1 (bottom right corner).
            color (Tuple[int, int, int]): color of the outline of the box.

        Returns:
            np.ndarray: image object with box drawn.
        """
        cv2.rectangle(self.output, pt1, pt2, color, thickness=self.line_width, lineType=cv2.LINE_AA)
        return self.output

    @_dependency.requires_module("cv2")
    def draw_text(
        self,
        text: str,
        pt1: Tuple[int, int],
        pt2: Tuple[int, int],
        *,
        font_size: Optional[int] = None,
        color: Tuple[int, int, int] = (229, 160, 21),
        txt_color: Tuple[int, int, int] = (255, 255, 255),
    ):
        """
        Draws text on given image.

        Args:
            text (string): class label
            pt1 (Tuple[int, int]): Vertex of the rectangle (top left corner).
            pt2 (Tuple[int, int]): Vertex of the rectangle opposite to pt1 (bottom right corner).
            font_size (int, optional): font of the text. If not provided, a font size
                proportional to the image width is calculated and used. Default: None
            color (Tuple[int, int, int]): color of the filled text.
            txt_color (Tuple[int, int, int]): color of the text.

        Returns:
            np.ndarray: image object with text drawn.
        """

        if font_size is None:
            font_size = max(self.line_width - 1, 1)  # font thickness
        w, h = cv2.getTextSize(text, 0, fontScale=self.line_width / 3, thickness=font_size)[0]
        outside = pt1[1] - h - 3 >= 0  # text fits outside box
        pt2 = pt1[0] + w, pt1[1] - h - 3 if outside else pt1[1] + h + 3
        cv2.rectangle(self.output, pt1, pt2, color, -1, cv2.LINE_AA)  # filled
        cv2.putText(
            self.output,
            text,
            (pt1[0], pt1[1] - 2 if outside else pt1[1] + h + 2),
            0,
            self.line_width / 3,
            txt_color,
            thickness=font_size,
            lineType=cv2.LINE_AA,
        )
        return self.output

    def _convert_boxes(self, boxes: Union[Tensor, np.ndarray]):
        """
        Convert different format of boxes to an Nx4 array.
        """
        if isinstance(boxes, Tensor):
            return boxes.cpu().detach().numpy()
        else:
            return boxes

    def _create_text_labels(
        self,
        classes: Optional[List[int]] = None,
        scores: Optional[List[float]] = None,
        is_crowd: Optional[List[bool]] = None,
    ):
        """
        Generate labels that classes and scores can match, and set class back to its original
        name if concrete class names are provided.
        """
        labels = None
        if classes is not None:
            if self.metadata is not None and len(self.metadata) > 0:
                labels = [self.metadata[i] for i in classes]
            else:
                labels = [str(i) for i in classes]
        if scores is not None:
            if labels is None:
                labels = [f"{score * 100:.0f}%" for score in scores]
            else:
                labels = [f"{label} {score * 100:.0f}%" for label, score in zip(labels, scores)]
        if labels is not None and is_crowd is not None:
            labels = [label + ("|crowd" if crowd else "") for label, crowd in zip(labels, is_crowd)]
        return labels

    def _create_colors(self, labels: Optional[List[int]] = None):
        """
        Generate colors that match the labels.
        """
        colors = None
        if labels is not None:
            colors = [self.assigned_colors(label, bgr=self.is_bgr) for label in labels]
        return colors

    def _change_color_brightness(
        self,
        color: Tuple[int, int, int],
        brightness_factor: float,
    ) -> Tuple[int, int, int]:
        """
        Depending on the brightness_factor, gives a lighter or darker color i.e. a color with
        less or more saturation than the original color.

        Args:
            color (Tuple[int, int, int]): color of the polygon.
            brightness_factor (float): a value in [-1.0, 1.0] range. A lightness factor of
                0 will correspond to no change, a factor in [-1.0, 0) range will result in
                a darker color and a factor in (0, 1.0] range will result in a lighter color.

        Returns:
            modified_color (Tuple[int, int, int]): a tuple containing the RGB/BGR values of the
                modified color.
        """
        assert brightness_factor >= -1.0 and brightness_factor <= 1.0
        # TODO: Implement the details in a follow-up PR
        return color



# Copyright (c) 2021, Zhiqiang Wang. All Rights Reserved.
import hashlib

from torch import Tensor
from yolort.v5 import AutoShape, attempt_download, load_yolov5_model


def test_attempt_download():

    model_url = "https://github.com/ultralytics/yolov5/releases/download/v4.0/yolov5s.pt"
    checkpoint_path = attempt_download(model_url, hash_prefix="9ca9a642")
    with open(checkpoint_path, "rb") as f:
        bytes = f.read()  # read entire file as bytes
        readable_hash = hashlib.sha256(bytes).hexdigest()
    assert readable_hash[:8] == "9ca9a642"


def test_load_yolov5_model_autoshape_attached():
    img_path = "test/assets/zidane.jpg"

    model_url = "https://github.com/ultralytics/yolov5/releases/download/v6.1/yolov5s.pt"
    checkpoint_path = attempt_download(model_url, hash_prefix="8b3b748c")

    model = load_yolov5_model(checkpoint_path)
    # Attach AutoShape
    model = AutoShape(model)

    results = model(img_path)

    assert isinstance(results.pred, list)
    assert len(results.pred) == 1
    assert isinstance(results.pred[0], Tensor)
    assert results.pred[0].shape == (4, 6)



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
Activation functions
"""

import torch
import torch.nn as nn
import torch.nn.functional as F


class SiLU(nn.Module):
    """
    Export-friendly version of nn.SiLU(). Starting with PyTorch 1.8,
    this operator supports exporting to ONNX, and there is also a
    build-in implementation of it on TVM.

    Ref: <https://arxiv.org/pdf/1606.08415.pdf>
    """

    @staticmethod
    def forward(x):
        return x * torch.sigmoid(x)


class Hardswish(nn.Module):
    """
    Export-friendly version of nn.Hardswish(). Starting with PyTorch 1.8,
    this operator supports exporting to ONNX, and currently this module
    is only used for TVM.
    """

    @staticmethod
    def forward(x):
        # return x * F.hardsigmoid(x)
        return x * F.hardtanh(x + 3, 0.0, 6.0) / 6.0



# Copyright (c) 2021, Zhiqiang Wang. All Rights Reserved.
from pathlib import Path

import numpy as np
import torch
from torch import Tensor
from yolort.data import _helper as data_helper
from yolort.data.data_module import DetectionDataModule
from yolort.utils import contains_any_tensor


def test_contains_any_tensor():
    dummy_numpy = np.random.randn(3, 6)
    assert not contains_any_tensor(dummy_numpy)
    dummy_tensor = torch.rand(3, 6)
    assert contains_any_tensor(dummy_tensor)
    dummy_tensors = [torch.rand(3, 6), torch.rand(9, 5)]
    assert contains_any_tensor(dummy_tensors)


def test_get_dataset():
    # Acquire the images and labels from the coco128 dataset
    train_dataset = data_helper.get_dataset(data_root="data-bin", mode="train")
    # Test the datasets
    image, target = next(iter(train_dataset))
    assert isinstance(image, Tensor)
    assert isinstance(target, dict)


def test_get_dataloader():
    batch_size = 8
    data_loader = data_helper.get_dataloader(data_root="data-bin", mode="train", batch_size=batch_size)
    # Test the dataloader
    images, targets = next(iter(data_loader))

    assert len(images) == batch_size
    assert isinstance(images[0], Tensor)
    assert len(images[0]) == 3
    assert len(targets) == batch_size
    assert isinstance(targets[0], dict)
    assert isinstance(targets[0]["image_id"], Tensor)
    assert isinstance(targets[0]["boxes"], Tensor)
    assert isinstance(targets[0]["labels"], Tensor)
    assert isinstance(targets[0]["orig_size"], Tensor)


def test_detection_data_module():
    # Setup the DataModule
    batch_size = 4
    train_dataset = data_helper.get_dataset(data_root="data-bin", mode="train")
    data_module = DetectionDataModule(train_dataset, batch_size=batch_size)
    assert data_module.batch_size == batch_size

    data_loader = data_module.train_dataloader()
    images, targets = next(iter(data_loader))
    assert len(images) == batch_size
    assert isinstance(images[0], Tensor)
    assert len(images[0]) == 3
    assert len(targets) == batch_size
    assert isinstance(targets[0], dict)
    assert isinstance(targets[0]["image_id"], Tensor)
    assert isinstance(targets[0]["boxes"], Tensor)
    assert isinstance(targets[0]["labels"], Tensor)


def test_prepare_coco128():
    data_path = Path("data-bin")
    coco128_dirname = "coco128"
    data_helper.prepare_coco128(data_path, dirname=coco128_dirname)
    annotation_file = data_path / coco128_dirname / "annotations" / "instances_train2017.json"
    assert annotation_file.is_file()



# Copyright (c) 2021, yolort team. All rights reserved.

from typing import Dict, List, Callable, Optional

import torch
from torch import nn, Tensor
from yolort.v5 import Conv, BottleneckCSP, C3, SPP


class IntermediateLevelP6(nn.Module):
    """
    This module is used to generate intermediate P6 block to the PAN.

    Args:
        x (List[Tensor]): the original feature maps

    Returns:
        results (List[Tensor]): the extended set of results
            of the PAN
    """

    def __init__(
        self,
        depth_multiple: float,
        in_channel: int,
        out_channel: int,
        version: str = "r4.0",
    ):
        super().__init__()

        block = _block[version]
        depth_gain = max(round(3 * depth_multiple), 1)

        self.p6 = nn.Sequential(
            Conv(in_channel, out_channel, k=3, s=2, version=version),
            block(out_channel, out_channel, n=depth_gain),
        )

    def forward(self, x: List[Tensor]) -> List[Tensor]:
        x.append(self.p6(x[-1]))
        return x


class PathAggregationNetwork(nn.Module):
    """
    Module that adds a PAN from on top of a set of feature maps. This is based on
    `"Path Aggregation Network for Instance Segmentation" <https://arxiv.org/abs/1803.01534>`_.

    The feature maps are currently supposed to be in increasing depth
    order.

    The input to the model is expected to be an OrderedDict[Tensor], containing
    the feature maps on top of which the PAN will be added.

    Args:
        in_channels (list[int]): number of channels for each feature map that
            is passed to the module
        out_channels (int): number of channels of the PAN representation
        version (str): ultralytics release version: ["r3.1", "r4.0", "r6.0"]

    Examples:

        >>> m = PathAggregationNetwork()
        >>> # get some dummy data
        >>> x = OrderedDict()
        >>> x['feat0'] = torch.rand(1, 128, 52, 44)
        >>> x['feat2'] = torch.rand(1, 256, 26, 22)
        >>> x['feat3'] = torch.rand(1, 512, 13, 11)
        >>> # compute the PAN on top of x
        >>> output = m(x)
        >>> print([(k, v.shape) for k, v in output.items()])
        [('feat0', torch.Size([1, 128, 52, 44])),
         ('feat2', torch.Size([1, 256, 26, 22])),
         ('feat3', torch.Size([1, 512, 13, 11]))]
    """

    def __init__(
        self,
        in_channels: List[int],
        depth_multiple: float,
        version: str = "r4.0",
        block: Optional[Callable[..., nn.Module]] = None,
        use_p6: bool = False,
    ):
        super().__init__()

        module_version = "r4.0" if version == "r6.0" else version

        # Define the Intermediate Block if necessary
        if use_p6:
            assert len(in_channels) == 4, "Length of in channels should be 4."
            intermediate_blocks = IntermediateLevelP6(
                depth_multiple,
                in_channels[2],
                in_channels[3],
                version=module_version,
            )
        else:
            assert len(in_channels) == 3, "Length of in channels should be 3."
            intermediate_blocks = None

        self.intermediate_blocks = intermediate_blocks

        if block is None:
            block = _block[module_version]

        depth_gain = max(round(3 * depth_multiple), 1)

        if version == "r6.0":
            init_block = SPP(in_channels[-1], in_channels[-1], k=(5, 9, 13))
        elif version in ["r3.1", "r4.0"]:
            init_block = block(in_channels[-1], in_channels[-1], n=depth_gain, shortcut=False)
        else:
            raise NotImplementedError(f"Version {version} is not implemented yet.")

        # Define the inner blocks
        inner_blocks = [init_block]

        if use_p6:
            in_channel = in_channels[1] + in_channels[-1]
            inner_blocks_p6 = [
                Conv(in_channels[-1], in_channels[2], 1, 1, version=module_version),
                nn.Upsample(scale_factor=2),
                block(in_channel, in_channels[2], n=depth_gain, shortcut=False),
            ]
            inner_blocks.extend(inner_blocks_p6)

        inner_blocks.extend(
            [
                Conv(in_channels[2], in_channels[1], 1, 1, version=module_version),
                nn.Upsample(scale_factor=2),
                block(in_channels[-1], in_channels[1], n=depth_gain, shortcut=False),
                Conv(in_channels[1], in_channels[0], 1, 1, version=module_version),
                nn.Upsample(scale_factor=2),
            ]
        )
        self.inner_blocks = nn.ModuleList(inner_blocks)

        # Define the layer blocks
        layer_blocks = [
            block(in_channels[1], in_channels[0], n=depth_gain, shortcut=False),
            Conv(in_channels[0], in_channels[0], 3, 2, version=module_version),
            block(in_channels[1], in_channels[1], n=depth_gain, shortcut=False),
            Conv(in_channels[1], in_channels[1], 3, 2, version=module_version),
            block(in_channels[-1], in_channels[2], n=depth_gain, shortcut=False),
        ]

        if use_p6:
            in_channel = in_channels[1] + in_channels[-1]
            layer_blocks_p6 = [
                Conv(in_channels[2], in_channels[2], 3, 2, version=module_version),
                block(in_channel, in_channels[-1], n=depth_gain, shortcut=False),
            ]
            layer_blocks.extend(layer_blocks_p6)

        self.layer_blocks = nn.ModuleList(layer_blocks)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                pass  # nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                m.eps = 1e-3
                m.momentum = 0.03
            elif isinstance(m, (nn.Hardswish, nn.LeakyReLU, nn.ReLU, nn.ReLU6)):
                m.inplace = True

    def get_result_from_inner_blocks(self, x: Tensor, idx: int) -> Tensor:
        """
        This is equivalent to self.inner_blocks[idx](x),
        but torchscript doesn't support this yet
        """
        num_blocks = len(self.inner_blocks)
        if idx < 0:
            idx += num_blocks
        i = 0
        out = x
        for module in self.inner_blocks:
            if i == idx:
                out = module(x)
            i += 1
        return out

    def get_result_from_layer_blocks(self, x: Tensor, idx: int) -> Tensor:
        """
        This is equivalent to self.layer_blocks[idx](x),
        but torchscript doesn't support this yet
        """
        num_blocks = len(self.layer_blocks)
        if idx < 0:
            idx += num_blocks
        i = 0
        out = x
        for module in self.layer_blocks:
            if i == idx:
                out = module(x)
            i += 1
        return out

    def forward(self, x: Dict[str, Tensor]) -> List[Tensor]:
        """
        Computes the PAN for a set of feature maps.

        Args:
            x (OrderedDict[Tensor]): feature maps for each feature level.

        Returns:
            results (OrderedDict[Tensor]): feature maps after PAN layers.
                They are ordered from highest resolution first.
        """
        # unpack OrderedDict into two lists for easier handling
        x = list(x.values())
        if self.intermediate_blocks is not None:
            x = self.intermediate_blocks(x)

        # Descending the feature pyramid
        num_features = len(x)
        inners = []
        last_inner = x[-1]
        for idx in range(num_features - 1):
            last_inner = self.get_result_from_inner_blocks(last_inner, 3 * idx)
            last_inner = self.get_result_from_inner_blocks(last_inner, 3 * idx + 1)
            inners.insert(0, last_inner)
            last_inner = self.get_result_from_inner_blocks(last_inner, 3 * idx + 2)
            last_inner = torch.cat([last_inner, x[num_features - idx - 2]], dim=1)

        inners.insert(0, last_inner)

        # Ascending the feature pyramid
        results = []
        last_inner = self.get_result_from_layer_blocks(inners[0], 0)
        results.append(last_inner)

        for idx in range(num_features - 1):
            last_inner = self.get_result_from_layer_blocks(last_inner, 2 * idx + 1)
            last_inner = torch.cat([last_inner, inners[idx + 1]], dim=1)
            last_inner = self.get_result_from_layer_blocks(last_inner, 2 * idx + 2)
            results.append(last_inner)

        return results


_block = {
    "r3.1": BottleneckCSP,
    "r4.0": C3,
}



# Copyright (c) 2021, yolort team. All rights reserved.

from pathlib import Path

import pytest
import pytorch_lightning as pl
from yolort.data import _helper as data_helper
from yolort.data.data_module import DetectionDataModule
from yolort.trainer import DefaultTask


def test_training_step():
    # Setup the DataModule
    data_path = "data-bin"
    train_dataset = data_helper.get_dataset(data_root=data_path, mode="train")
    val_dataset = data_helper.get_dataset(data_root=data_path, mode="val")
    data_module = DetectionDataModule(train_dataset, val_dataset, batch_size=8)
    # Load model
    model = DefaultTask(arch="yolov5n")
    model = model.train()
    # Trainer
    trainer = pl.Trainer(max_epochs=1)
    trainer.fit(model, data_module)


@pytest.mark.parametrize("arch, version, map5095, map50", [("yolov5s", "r4.0", 42.5, 65.3)])
def test_test_epoch_end(arch, version, map5095, map50):
    # Acquire the annotation file
    data_path = Path("data-bin")
    coco128_dirname = "coco128"
    data_helper.prepare_coco128(data_path, dirname=coco128_dirname)
    annotation_file = data_path / coco128_dirname / "annotations" / "instances_train2017.json"

    # Get dataloader to test
    val_dataloader = data_helper.get_dataloader(data_root=data_path, mode="val")

    # Load model
    model = DefaultTask(arch=arch, version=version, pretrained=True, annotation_path=annotation_file)

    # test step
    trainer = pl.Trainer(max_epochs=1)
    trainer.test(model, dataloaders=val_dataloader)
    # test epoch end
    results = model.evaluator.compute()
    assert results["AP"] > map5095
    assert results["AP50"] > map50



import contextlib
import io
import os
import warnings

import pytest
import torch
from torch import Tensor
from yolort import models
from yolort.models import YOLO, YOLOv5
from yolort.models.anchor_utils import AnchorGenerator
from yolort.models.backbone_utils import darknet_pan_backbone
from yolort.models.box_head import YOLOHead, PostProcess, SetCriterion
from yolort.models.transformer import darknet_tan_backbone
from yolort.models.yolo_lite import yolov5_mobilenet_v3_small_fpn
from yolort.v5 import get_yolov5_size, attempt_download


@contextlib.contextmanager
def freeze_rng_state():
    rng_state = torch.get_rng_state()
    if torch.cuda.is_available():
        cuda_rng_state = torch.cuda.get_rng_state()
    yield
    if torch.cuda.is_available():
        torch.cuda.set_rng_state(cuda_rng_state)
    torch.set_rng_state(rng_state)


def _check_jit_scriptable(nn_module, args, unwrapper=None, skip=False):
    """
    Check that a nn.Module's results in TorchScript match eager and that it can be exported
    https://github.com/pytorch/vision/blob/12fd3a6/test/test_models.py#L90-L141
    """

    def assert_export_import_module(m, args):
        """
        Check that the results of a model are the same after saving and loading
        """

        def get_export_import_copy(m):
            """
            Save and load a TorchScript model
            """
            buffer = io.BytesIO()
            torch.jit.save(m, buffer)
            buffer.seek(0)
            imported = torch.jit.load(buffer)
            return imported

        m_import = get_export_import_copy(m)
        with freeze_rng_state():
            results = m(*args)
        with freeze_rng_state():
            results_from_imported = m_import(*args)
        tol = 3e-4
        try:
            torch.testing.assert_close(results, results_from_imported, atol=tol, rtol=tol)
        except ValueError:
            # custom check for the models that return named tuples:
            # we compare field by field while ignoring None as assert_close can't handle None
            for a, b in zip(results, results_from_imported):
                if a is not None:
                    torch.testing.assert_close(a, b, atol=tol, rtol=tol)

    TEST_WITH_SLOW = os.getenv("PYTORCH_TEST_WITH_SLOW", "0") == "1"
    if not TEST_WITH_SLOW or skip:
        # TorchScript is not enabled, skip these tests
        msg = (
            f"The check_jit_scriptable test for {nn_module.__class__.__name__} "
            "was skipped. This test checks if the module's results in TorchScript "
            "match eager and that it can be exported. To run these tests make "
            "sure you set the environment variable PYTORCH_TEST_WITH_SLOW=1 and "
            "that the test is not manually skipped."
        )
        warnings.warn(msg, RuntimeWarning)
        return None

    sm = torch.jit.script(nn_module)

    with freeze_rng_state():
        eager_out = nn_module(*args)

    with freeze_rng_state():
        script_out = sm(*args)
        if unwrapper:
            script_out = unwrapper(script_out)

    torch.testing.assert_close(eager_out, script_out, atol=1e-4, rtol=1e-4)
    assert_export_import_module(sm, args)


class TestModel:

    num_classes = 80
    num_outputs = num_classes + 5

    @staticmethod
    def _get_in_channels(width_multiple, use_p6):
        grow_widths = [256, 512, 768, 1024] if use_p6 else [256, 512, 1024]
        in_channels = [int(gw * width_multiple) for gw in grow_widths]
        return in_channels

    @staticmethod
    def _get_strides(use_p6: bool):
        if use_p6:
            strides = [8, 16, 32, 64]
        else:
            strides = [8, 16, 32]
        return strides

    @staticmethod
    def _get_anchor_grids(use_p6: bool):
        if use_p6:
            anchor_grids = [
                [19, 27, 44, 40, 38, 94],
                [96, 68, 86, 152, 180, 137],
                [140, 301, 303, 264, 238, 542],
                [436, 615, 739, 380, 925, 792],
            ]
        else:
            anchor_grids = [
                [10, 13, 16, 30, 33, 23],
                [30, 61, 62, 45, 59, 119],
                [116, 90, 156, 198, 373, 326],
            ]
        return anchor_grids

    def _compute_anchors(self, height, width, use_p6: bool):
        strides = self._get_strides(use_p6)
        anchors_num = len(strides)
        anchors_shape = []
        for s in strides:
            anchors_shape.append((height // s, width // s))
        return anchors_num, anchors_shape

    def _get_feature_shapes(self, height, width, width_multiple=0.5, use_p6=False):
        in_channels = self._get_in_channels(width_multiple, use_p6)
        strides = self._get_strides(use_p6)

        return [(c, height // s, width // s) for (c, s) in zip(in_channels, strides)]

    def _get_feature_maps(self, batch_size, height, width, width_multiple=0.5, use_p6=False):
        feature_shapes = self._get_feature_shapes(
            height,
            width,
            width_multiple=width_multiple,
            use_p6=use_p6,
        )
        feature_maps = [torch.rand(batch_size, *f_shape) for f_shape in feature_shapes]
        return feature_maps

    def _get_head_outputs(self, batch_size, height, width, width_multiple=0.5, use_p6=False):
        feature_shapes = self._get_feature_shapes(
            height,
            width,
            width_multiple=width_multiple,
            use_p6=use_p6,
        )

        num_outputs = self.num_outputs
        head_shapes = [(batch_size, 3, *f_shape[1:], num_outputs) for f_shape in feature_shapes]
        head_outputs = [torch.rand(*h_shape) for h_shape in head_shapes]

        return head_outputs

    def _init_test_backbone_with_pan(
        self,
        depth_multiple,
        width_multiple,
        version,
        use_p6,
        use_tan,
    ):
        model_size = get_yolov5_size(depth_multiple, width_multiple)
        backbone_name = f"darknet_{model_size}_{version.replace('.', '_')}"
        backbone_arch = eval(f"darknet_{'tan' if use_tan else 'pan'}_backbone")
        assert backbone_arch in [darknet_pan_backbone, darknet_tan_backbone]
        model = backbone_arch(
            backbone_name,
            depth_multiple,
            width_multiple,
            version=version,
            use_p6=use_p6,
        )
        return model

    @pytest.mark.parametrize(
        "depth_multiple, width_multiple, version, use_p6, use_tan",
        [
            (0.33, 0.5, "r4.0", False, True),
            (0.33, 0.5, "r3.1", False, False),
            (0.33, 0.5, "r4.0", False, False),
            (0.33, 0.5, "r6.0", False, False),
            (0.33, 0.5, "r6.0", True, False),
            (0.67, 0.75, "r6.0", False, False),
        ],
    )
    @pytest.mark.parametrize("batch_size, height, width", [(4, 448, 320), (2, 384, 640)])
    def test_backbone_with_pan(
        self,
        depth_multiple,
        width_multiple,
        version,
        use_p6,
        use_tan,
        batch_size,
        height,
        width,
    ):
        out_shape = self._get_feature_shapes(height, width, width_multiple=width_multiple, use_p6=use_p6)

        x = torch.rand(batch_size, 3, height, width)
        model = self._init_test_backbone_with_pan(
            depth_multiple, width_multiple, version, use_p6, use_tan=use_tan
        )
        out = model(x)

        expected_num_output = 4 if use_p6 else 3
        assert len(out) == expected_num_output
        for i in range(expected_num_output):
            assert tuple(out[i].shape) == (batch_size, *out_shape[i])

        _check_jit_scriptable(model, (x,))

    def _init_test_anchor_generator(self, use_p6=False):
        strides = self._get_strides(use_p6)
        anchor_grids = self._get_anchor_grids(use_p6)
        anchor_generator = AnchorGenerator(strides, anchor_grids)
        return anchor_generator

    @pytest.mark.parametrize(
        "width_multiple, use_p6",
        [(0.5, False), (0.5, True)],
    )
    @pytest.mark.parametrize("batch_size, height, width", [(4, 448, 320), (2, 384, 640)])
    def test_anchor_generator(self, width_multiple, use_p6, batch_size, height, width):
        feature_maps = self._get_feature_maps(
            batch_size, height, width, width_multiple=width_multiple, use_p6=use_p6
        )
        model = self._init_test_anchor_generator(use_p6)
        anchors = model(feature_maps)
        expected_anchors_num, expected_anchors_shape = self._compute_anchors(height, width, use_p6)

        assert len(anchors) == 2
        assert len(anchors[0]) == len(anchors[1]) == expected_anchors_num
        for i in range(expected_anchors_num):
            assert tuple(anchors[0][i].shape) == (1, 3, *(expected_anchors_shape[i]), 2)
            assert tuple(anchors[1][i].shape) == (1, 3, *(expected_anchors_shape[i]), 2)

        _check_jit_scriptable(model, (feature_maps,))

    def _init_test_yolo_head(self, width_multiple=0.5, use_p6=False):
        in_channels = self._get_in_channels(width_multiple, use_p6)
        strides = self._get_strides(use_p6)
        num_anchors = len(strides)
        num_classes = self.num_classes

        box_head = YOLOHead(in_channels, num_anchors, strides, num_classes)
        return box_head

    def test_yolo_head(self):
        N, H, W = 4, 416, 352
        feature_maps = self._get_feature_maps(N, H, W)
        model = self._init_test_yolo_head()
        head_outputs = model(feature_maps)
        assert len(head_outputs) == 3

        target_head_outputs = self._get_head_outputs(N, H, W)

        assert head_outputs[0].shape == target_head_outputs[0].shape
        assert head_outputs[1].shape == target_head_outputs[1].shape
        assert head_outputs[2].shape == target_head_outputs[2].shape
        _check_jit_scriptable(model, (feature_maps,))

    def _init_test_postprocessors(self, strides):
        score_thresh = 0.5
        nms_thresh = 0.45
        detections_per_img = 100
        postprocessors = PostProcess(strides, score_thresh, nms_thresh, detections_per_img)
        return postprocessors

    @pytest.mark.parametrize("use_p6", [False, True])
    def test_postprocessors(self, use_p6):
        N, H, W = 4, 416, 352
        strides = self._get_strides(use_p6)
        feature_maps = self._get_feature_maps(N, H, W, use_p6=use_p6)
        head_outputs = self._get_head_outputs(N, H, W, use_p6=use_p6)

        anchor_generator = self._init_test_anchor_generator(use_p6=use_p6)
        grids, shifts = anchor_generator(feature_maps)
        model = self._init_test_postprocessors(strides)
        out = model(head_outputs, grids, shifts)

        assert len(out) == N
        assert isinstance(out[0], dict)
        assert isinstance(out[0]["boxes"], Tensor)
        assert isinstance(out[0]["labels"], Tensor)
        assert isinstance(out[0]["scores"], Tensor)
        _check_jit_scriptable(model, (head_outputs, grids, shifts))

    def test_criterion(self, use_p6=False):
        N, H, W = 4, 640, 640
        head_outputs = self._get_head_outputs(N, H, W)
        strides = self._get_strides(use_p6)
        anchor_grids = self._get_anchor_grids(use_p6)
        num_classes = self.num_classes

        targets = torch.tensor(
            [
                [0.0000, 7.0000, 0.0714, 0.3749, 0.0760, 0.0654],
                [0.0000, 1.0000, 0.1027, 0.4402, 0.2053, 0.1920],
                [1.0000, 5.0000, 0.4720, 0.6720, 0.3280, 0.1760],
                [3.0000, 3.0000, 0.6305, 0.3290, 0.3274, 0.2270],
            ]
        )
        criterion = SetCriterion(strides, anchor_grids, num_classes)
        losses = criterion(targets, head_outputs)
        assert isinstance(losses, dict)
        assert isinstance(losses["cls_logits"], Tensor)
        assert isinstance(losses["bbox_regression"], Tensor)
        assert isinstance(losses["objectness"], Tensor)


@pytest.mark.parametrize("arch", ["yolov5s", "yolov5m", "yolov5l", "yolov5ts"])
def test_torchscript(arch):
    model = models.__dict__[arch](pretrained=True, size=(320, 320), score_thresh=0.45)
    model.eval()

    scripted_model = torch.jit.script(model)
    scripted_model.eval()

    x = [torch.rand(3, 288, 320), torch.rand(3, 300, 256)]

    out = model(x)
    out_script = scripted_model(x)

    torch.testing.assert_close(out[0]["scores"], out_script[1][0]["scores"], rtol=0, atol=0)
    torch.testing.assert_close(out[0]["labels"], out_script[1][0]["labels"], rtol=0, atol=0)
    torch.testing.assert_close(out[0]["boxes"], out_script[1][0]["boxes"], rtol=0, atol=0)


@pytest.mark.parametrize(
    "arch, size_divisible, version, upstream_version, hash_prefix",
    [
        ("yolov5s", 32, "r4.0", "v4.0", "9ca9a642"),
        ("yolov5n", 32, "r6.0", "v6.0", "649e089f"),
        ("yolov5s", 32, "r6.0", "v6.0", "c3b140f3"),
        ("yolov5n6", 64, "r6.0", "v6.0", "beecbbae"),
    ],
)
def test_load_from_yolov5(arch, size_divisible, version, upstream_version, hash_prefix):
    img_path = "test/assets/bus.jpg"

    base_url = "https://github.com/ultralytics/yolov5/releases/download/"
    model_url = f"{base_url}/{upstream_version}/{arch}.pt"
    checkpoint_path = attempt_download(model_url, hash_prefix=hash_prefix)

    score_thresh = 0.25

    model_yolov5 = YOLOv5.load_from_yolov5(
        checkpoint_path,
        score_thresh=score_thresh,
        version=version,
        size_divisible=size_divisible,
    )
    model_yolov5.eval()
    out_from_yolov5 = model_yolov5.predict(img_path)
    assert isinstance(out_from_yolov5[0], dict)
    assert isinstance(out_from_yolov5[0]["boxes"], Tensor)
    assert isinstance(out_from_yolov5[0]["labels"], Tensor)
    assert isinstance(out_from_yolov5[0]["scores"], Tensor)

    model = models.__dict__[arch](
        upstream_version=version,
        pretrained=True,
        score_thresh=score_thresh,
    )
    model.eval()
    out = model.predict(img_path)

    torch.testing.assert_close(out_from_yolov5[0]["scores"], out[0]["scores"], rtol=0, atol=0)
    torch.testing.assert_close(out_from_yolov5[0]["labels"], out[0]["labels"], rtol=0, atol=0)
    torch.testing.assert_close(out_from_yolov5[0]["boxes"], out[0]["boxes"], rtol=0, atol=0)


@pytest.mark.parametrize(
    "arch, size_divisible, version, upstream_version, hash_prefix",
    [
        ("yolov5s", 32, "r4.0", "v4.0", "9ca9a642"),
        ("yolov5n", 32, "r6.0", "v6.0", "649e089f"),
        ("yolov5s", 32, "r6.0", "v6.0", "c3b140f3"),
        ("yolov5n6", 64, "r6.0", "v6.0", "beecbbae"),
    ],
)
def test_load_from_yolov5_torchscript(arch, size_divisible, version, upstream_version, hash_prefix):
    import cv2
    from yolort.utils import read_image_to_tensor
    from yolort.v5 import letterbox

    # Loading and pre-processing the image
    img_path = "test/assets/zidane.jpg"
    img_raw = cv2.imread(img_path)
    img = letterbox(img_raw, new_shape=(640, 640), stride=size_divisible)[0]
    img = read_image_to_tensor(img)

    base_url = "https://github.com/ultralytics/yolov5/releases/download/"
    model_url = f"{base_url}/{upstream_version}/{arch}.pt"
    checkpoint_path = attempt_download(model_url, hash_prefix=hash_prefix)

    score_thresh = 0.25

    model = YOLO.load_from_yolov5(checkpoint_path, score_thresh=score_thresh, version=version)
    model.eval()
    scripted_model = torch.jit.script(model)
    scripted_model.eval()

    out = model(img[None])
    out_script = scripted_model(img[None])

    torch.testing.assert_close(out[0]["scores"], out_script[1][0]["scores"], rtol=0, atol=0)
    torch.testing.assert_close(out[0]["labels"], out_script[1][0]["labels"], rtol=0, atol=0)
    torch.testing.assert_close(out[0]["boxes"], out_script[1][0]["boxes"], rtol=0, atol=0)


def test_yolov5_mobilenet_v3_small_fpn():

    model = yolov5_mobilenet_v3_small_fpn()
    model = model.eval()

    images = torch.rand(4, 3, 320, 320)
    out = model(images)
    assert isinstance(out, list)
    assert len(out) == 4
    assert isinstance(out[0], dict)
    assert isinstance(out[0]["boxes"], Tensor)
    assert isinstance(out[0]["labels"], Tensor)
    assert isinstance(out[0]["scores"], Tensor)



# Copyright (c) 2021, yolort team. All rights reserved.

import argparse
from pathlib import PosixPath
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import yolort.models as models
from pytorch_lightning import LightningModule
from torch import Tensor
from torchvision.ops import box_iou
from yolort.data.coco_eval import COCOEvaluator


__all__ = ["DefaultTask"]


def _evaluate_iou(target, pred):
    """
    Evaluate intersection over union (IOU) for target from dataset and
    output prediction from model
    """
    if pred["boxes"].shape[0] == 0:
        # no box detected, 0 IOU
        return torch.tensor(0.0, device=pred["boxes"].device)
    return box_iou(target["boxes"], pred["boxes"]).diag().mean()


class DefaultTask(LightningModule):
    """
    Wrapping the trainer into the YOLOv5 Module.

    Args:
        arch (string): YOLOv5 model architecture. Default: 'yolov5s'
        version (str): model released by the upstream YOLOv5. Possible values
            are ['r6.0']. Default: 'r6.0'.
        lr (float): The initial learning rate
        annotation_path (Optional[Union[string, PosixPath]]): Path of the COCO annotation file
            Default: None.
    """

    def __init__(
        self,
        arch: str = "yolov5s",
        version: str = "r6.0",
        lr: float = 0.01,
        annotation_path: Optional[Union[str, PosixPath]] = None,
        **kwargs: Any,
    ) -> None:

        super().__init__()

        self.model = models.__dict__[arch](upstream_version=version, **kwargs)
        self.lr = lr

        # evaluators for validation datasets
        self.evaluator = None
        if annotation_path is not None:
            self.evaluator = COCOEvaluator(annotation_path, iou_type="bbox")

        # used only on torchscript mode
        self._has_warned = False

    def forward(
        self,
        inputs: List[Tensor],
        targets: Optional[List[Dict[str, Tensor]]] = None,
    ) -> Tuple[Dict[str, Tensor], List[Dict[str, Tensor]]]:
        """
        This exists since PyTorchLightning forward are used for inference only (separate from
        ``training_step``). We keep ``targets`` here for Backward Compatible.
        """
        return self.model(inputs, targets)

    def training_step(self, batch, batch_idx):
        """
        The training step.
        """
        loss_dict = self.model(*batch)
        loss = sum(loss_dict.values())
        self.log_dict(loss_dict, on_step=True, on_epoch=True, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        images, targets = batch
        # fasterrcnn takes only images for eval() mode
        preds = self.model(images)
        iou = torch.stack([_evaluate_iou(t, o) for t, o in zip(targets, preds)]).mean()
        outs = {"val_iou": iou}
        self.log_dict(outs, on_step=True, on_epoch=True, prog_bar=True)
        return outs

    def validation_epoch_end(self, outs):
        avg_iou = torch.stack([o["val_iou"] for o in outs]).mean()
        self.log("avg_val_iou", avg_iou)

    def test_step(self, batch, batch_idx):
        """
        The test step.
        """
        images, targets = batch
        images = list(image.to(next(self.parameters()).device) for image in images)
        preds = self.model(images)
        results = self.evaluator(preds, targets)
        # log step metric
        self.log("eval_step", results, prog_bar=True, on_step=True)

    def test_epoch_end(self, outputs):
        return self.log("coco_eval", self.evaluator.compute())

    def configure_optimizers(self):
        return torch.optim.SGD(
            self.model.parameters(),
            lr=self.lr,
            momentum=0.9,
            weight_decay=5e-4,
        )

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = argparse.ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument("--arch", default="yolov5_darknet_pan_s_r40", help="model architecture")
        parser.add_argument(
            "--pretrained",
            action="store_true",
            help="Use pre-trained models from the modelzoo",
        )
        parser.add_argument(
            "--lr",
            default=0.01,
            type=float,
            help="initial learning rate, 0.01 is the default value for training "
            "on 8 gpus and 2 images_per_gpu",
        )
        parser.add_argument("--momentum", default=0.9, type=float, metavar="M", help="momentum")
        parser.add_argument(
            "--weight-decay",
            default=5e-4,
            type=float,
            metavar="W",
            help="weight decay (default: 5e-4)",
        )
        return parser
